package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** currently this class exports only static methods to create repetitions; there are no public instance methods or constructors */
public class Repeat extends Union {

    /** repeat zero or one times */
    public  static Repeat maybe(Element e)              { return new Repeat(e, true, false); }
    /** repeat zero or more times */
    public  static Repeat many0(Element e)              { return new Repeat(e, true, true); }
    /** repeat zero or more times, separated by <tt>sep</tt> */
    public  static Repeat many0(Element e, Element sep) { return new Repeat(e, true, true, sep); }
    /** repeat one or more times */
    public  static Repeat many1(Element e)              { return new Repeat(e, false, true); }
    /** repeat one or more times, separated by <tt>sep</tt> */
    public  static Repeat many1(Element e, Element sep) { return new Repeat(e, false, true, sep); }
    public  static Maximal maximal(Element e)           { return new Maximal(e); }

    // Instance //////////////////////////////////////////////////////////////////////////////
    
    final boolean zeroOkay, manyOkay;
    final Element e;
    final Element separator;

    Repeat(final Element e, boolean zeroOkay, boolean manyOkay) { this(e, zeroOkay, manyOkay, null); }
    Repeat(final Element e, boolean zeroOkay, boolean manyOkay, Element separator) {
        super(e+(!manyOkay ? "?" : (zeroOkay ? "*" : "+")), true);
        this.e = e;
        this.zeroOkay = zeroOkay;
        this.manyOkay = manyOkay;
        this.separator = separator;
        if (zeroOkay) {
            add(new Sequence.Constant.Empty());
            if (manyOkay) add(new Sequence.Singleton(many1(e, separator), null, null));
            else          add(new Sequence.Singleton(e, null, null));
        } else {
            add(new Sequence.RewritingSequence(null, new Element[] { e }, null, null));
            if (this.separator==null) {
                add(new Sequence.Unwrap(new Element[] { e,                 Repeat.this }, null, null));
            } else {
                add(new Sequence.Unwrap(new Element[] { e, this.separator, Repeat.this }, new boolean[] { false, true, false }, null, null));
            }
        }
    }

    static class MaximalSequence extends Sequence.Singleton {
        private final Element e;
        public String toString() { return e+"@";}
        public MaximalSequence(Element e) { super(e, null, null); this.e = e; }
    }
    static class Maximal extends Union {
        public Maximal(final Element e) {
            super(e+"+", true);
            add(new MaximalSequence(e));
        }
    }
}
